# -*- coding: utf-8 -*-
"""Generic exceptions for the Mantarray SDK."""
from typing import Any
from typing import Tuple

from nptyping import NDArray
import numpy as np


class UnrecognizedFilterUuidError(Exception):
    pass


class FilterCreationNotImplementedError(Exception):
    pass


class DataAlreadyLoadedInPipelineError(Exception):
    pass


class PeakDetectionError(Exception):
    pass


class InsufficientPeaksDetectedError(PeakDetectionError):
    pass


class TwoFeaturesInARowError(PeakDetectionError):
    """There must always be a peak in between valleys and vice-versa."""

    def __init__(
        self,
        peak_and_valley_timepoints: Tuple[NDArray[int], NDArray[int]],
        filtered_data: NDArray[(2, Any), int],
        back_to_back_points: Tuple[int, int],
        feature_name: str = "feature",
    ) -> None:
        peak_timepoints, valley_timepoints = peak_and_valley_timepoints
        with np.printoptions(threshold=np.inf):  # don't truncate the numpy array

            base_msg = f"Detected peak timepoints: {peak_timepoints}\nDetected valley timepoints: {valley_timepoints}\nData used in peak detection: {filtered_data}\n"
        prepend_msg = f"Two back-to-back {feature_name}s in a row were detected at timepoints: {back_to_back_points[0]} and {back_to_back_points[1]}\n"
        super().__init__(prepend_msg + base_msg)


class TwoValleysInARowError(TwoFeaturesInARowError):
    """There must always be a peak in between valleys."""

    def __init__(
        self,
        peak_and_valley_timepoints: Tuple[NDArray[int], NDArray[int]],
        filtered_data: NDArray[(2, Any), int],
        back_to_back_points: Tuple[int, int],
    ) -> None:
        super().__init__(
            peak_and_valley_timepoints,
            filtered_data,
            back_to_back_points,
            feature_name="valley",
        )


class TwoPeaksInARowError(TwoFeaturesInARowError):
    """There must always be a valley in between peaks."""

    def __init__(
        self,
        peak_and_valley_timepoints: Tuple[NDArray[int], NDArray[int]],
        filtered_data: NDArray[(2, Any), int],
        back_to_back_points: Tuple[int, int],
    ) -> None:
        super().__init__(
            peak_and_valley_timepoints,
            filtered_data,
            back_to_back_points,
            feature_name="peak",
        )


class TooFewPeaksDetectedError(PeakDetectionError):
    pass



# -*- coding: utf-8 -*-
"""Global constants."""
import uuid

CENTIMILLISECONDS_PER_SECOND = 100000
TWITCH_PERIOD_UUID = uuid.UUID("6e0cd81c-7861-4c49-ba14-87b2739d65fb")
TWITCH_FREQUENCY_UUID = uuid.UUID(
    "472d0707-ff87-4198-9374-c28900bb216c"
)  # This is just the reciprocal of twitch period, but is pre-computed to make downstream pipelines simpler. Frequency is reported in Hz
AMPLITUDE_UUID = uuid.UUID("89cf1105-a015-434f-b527-4169b9400e26")
AUC_UUID = uuid.UUID("e7b9a6e4-c43d-4e8b-af7e-51742e252030")
WIDTH_UUID = uuid.UUID("c4c60d55-017a-4783-9600-f19606de26f3")
WIDTH_VALUE_UUID = uuid.UUID("05041f4e-c77d-42d9-a2ae-8902f912e9ac")
WIDTH_RISING_COORDS_UUID = uuid.UUID("2a16acb6-4df7-4064-9d47-5d27ea7a98ad")
WIDTH_FALLING_COORDS_UUID = uuid.UUID("26e5637d-42c9-4060-aa5d-52209b349c84")

PRIOR_PEAK_INDEX_UUID = uuid.UUID("80df90dc-21f8-4cad-a164-89436909b30a")
PRIOR_VALLEY_INDEX_UUID = uuid.UUID("72ba9466-c203-41b6-ac30-337b4a17a124")
SUBSEQUENT_PEAK_INDEX_UUID = uuid.UUID("7e37325b-6681-4623-b192-39f154350f36")
SUBSEQUENT_VALLEY_INDEX_UUID = uuid.UUID("fd47ba6b-ee4d-4674-9a89-56e0db7f3d97")

BESSEL_BANDPASS_UUID = uuid.UUID("0ecf0e52-0a29-453f-a6ff-46f5ec3ae783")
BESSEL_LOWPASS_10_UUID = uuid.UUID("7d64cac3-b841-4912-b734-c0cf20a81e7a")
BESSEL_LOWPASS_30_UUID = uuid.UUID("eee66c75-4dc4-4eb4-8d48-6c608bf28d91")
BUTTERWORTH_LOWPASS_30_UUID = uuid.UUID("de8d8cef-65bf-4119-ada7-bdecbbaa897a")

# GMR conversion factors
MIDSCALE_CODE = 0x800000
RAW_TO_SIGNED_CONVERSION_VALUE = 2 ** 23  # subtract this value from raw hardware data

# REFERENCE_VOLTAGE = 3.3

# ADC_GAIN = 32

MIN_NUMBER_PEAKS = 3



# -*- coding: utf-8 -*-
"""Transforming arrays of Mantarray data throughout the analysis pipeline."""
from typing import Any
from typing import Dict
from typing import List
from typing import Union
import uuid

from nptyping import NDArray
import numpy as np
from scipy import signal

from .constants import BESSEL_BANDPASS_UUID
from .constants import BESSEL_LOWPASS_10_UUID
from .constants import BESSEL_LOWPASS_30_UUID
from .constants import BUTTERWORTH_LOWPASS_30_UUID
from .constants import CENTIMILLISECONDS_PER_SECOND
from .constants import RAW_TO_SIGNED_CONVERSION_VALUE
from .exceptions import FilterCreationNotImplementedError
from .exceptions import UnrecognizedFilterUuidError

FILTER_CHARACTERISTICS: Dict[uuid.UUID, Dict[str, Union[str, float, int]]] = {
    BESSEL_BANDPASS_UUID: {
        "filter_type": "bessel",
        "order": 4,
        "high_pass_hz": 0.1,
        "low_pass_hz": 10,
    },
    BESSEL_LOWPASS_10_UUID: {"filter_type": "bessel", "order": 4, "low_pass_hz": 10},
    BESSEL_LOWPASS_30_UUID: {"filter_type": "bessel", "order": 4, "low_pass_hz": 30},
    BUTTERWORTH_LOWPASS_30_UUID: {
        "filter_type": "butterworth",
        "order": 4,
        "low_pass_hz": 30,
    },
}


def create_filter(
    filter_uuid: uuid.UUID, sample_period_centimilliseconds: int,
) -> NDArray[(Any, Any), float]:
    """Create a filter to apply to data streams.

    Args:
        filter_uuid: a UUID of an already accepted and approved filter
        sample_period_centimilliseconds: the sampling period for the data stream you want to apply the filter to

    Returns:
        The scipy 'b' and 'a' vectors to use in scipy.signal.filtfilt
    """
    sampling_frequency_hz = 1 / (
        sample_period_centimilliseconds / CENTIMILLISECONDS_PER_SECOND
    )
    nyquist_frequency_limit = sampling_frequency_hz / 2

    if filter_uuid not in FILTER_CHARACTERISTICS:
        raise UnrecognizedFilterUuidError(filter_uuid)

    the_filter_characteristics = FILTER_CHARACTERISTICS[filter_uuid]
    normalized_high_pass_frequency: Union[int, float]
    normalized_low_pass_frequency: Union[int, float]
    filter_order: int
    if "order" in the_filter_characteristics:
        if not isinstance(the_filter_characteristics["order"], int):
            raise NotImplementedError("The filter order must always be an int.")

        filter_order = the_filter_characteristics["order"]
    pass_boundaries: List[Union[float, int]] = list()
    bandpass_type = "lowpass"  # by default
    if "high_pass_hz" in the_filter_characteristics:
        if isinstance(the_filter_characteristics["high_pass_hz"], str):
            raise NotImplementedError(
                "The high pass frequency should never be a string."
            )
        normalized_high_pass_frequency = (
            the_filter_characteristics["high_pass_hz"] / nyquist_frequency_limit
        )
        pass_boundaries.append(normalized_high_pass_frequency)
    if "low_pass_hz" in the_filter_characteristics:
        if isinstance(the_filter_characteristics["low_pass_hz"], str):
            raise NotImplementedError(
                "The low pass frequency should never be a string."
            )
        normalized_low_pass_frequency = (
            the_filter_characteristics["low_pass_hz"] / nyquist_frequency_limit
        )
        pass_boundaries.append(normalized_low_pass_frequency)

    if len(pass_boundaries) == 2:
        bandpass_type = "bandpass"
    filter_type = the_filter_characteristics["filter_type"]
    if filter_type == "bessel":
        sos_polys = signal.bessel(
            filter_order, pass_boundaries, btype=bandpass_type, output="sos"
        )
    elif filter_type == "butterworth":
        sos_polys = signal.butter(
            filter_order, pass_boundaries, btype=bandpass_type, output="sos"
        )
    else:
        raise FilterCreationNotImplementedError(filter_uuid)

    if not isinstance(sos_polys, NDArray[float]):
        raise NotImplementedError("Returned polynomials most be float arrays")
    return sos_polys


def apply_sensitivity_calibration(
    raw_gmr_reading: NDArray[(2, Any), int]
) -> NDArray[(2, Any), int]:
    """Apply the result of a sensor sensitivity calibration.

    Actual sensitivity calibration will be performed once information obtained from Jason.

    Args:
        raw_gmr_reading: an original 2d array of time vs GMR readings. Could be from tissue construct or reference sensor.

    Returns:
        A 2d array of the time vs GMR readings after sensitivity calibration. Data will be rounded to integers if calibration results in slight decimal behavior.
    """
    return raw_gmr_reading


def noise_cancellation(
    tissue_gmr_reading: NDArray[(2, Any), int],
    reference_gmr_reading: NDArray[
        (2, Any), int
    ],  # pylint: disable=unused-argument # this will eventually be used
) -> NDArray[(2, Any), int]:
    """Perform cancellation of ambient magnetic noise using reference sensor.

    This should be performed after sensitivity calibration has been applied to the raw data from each sensor.

    Args:
        tissue_gmr_reading: from the tissue construct sensor
        reference_gmr_reading: from the reference sensor

    Returns:
        A single 2D array of time vs GMR reading.
    """
    return tissue_gmr_reading


def apply_empty_plate_calibration(
    noise_cancelled_gmr: NDArray[(2, Any), int],
) -> NDArray[(2, Any), int]:
    """Apply the result of an empty plate calibration.

    Actual empty plate calibration will be performed once information obtained from Jason.

    Args:
        noise_cancelled_gmr: an 2D array of Time and GMR readings after combining reference and tissue sensor readings.

    Returns:
        A 2D array of the Time and GMR readings after empty plate calibration. Data will be rounded to integers if calibration results in slight decimal behavior.
    """
    return noise_cancelled_gmr


def apply_noise_filtering(
    fully_calibrated_gmr: NDArray[(2, Any), int],
    scipy_filter_sos_coefficients: NDArray[(Any, Any), float],
) -> NDArray[(2, Any), int]:
    """Apply the result of an empty plate calibration.

    Actual empty plate calibration will be performed once information obtained from Jason.

    Args:
        fully_calibrated_gmr: an 2D array of Time and GMR readings after applying the Empty Plate calibration.
        scipy_filter_sos_coefficients: The 'second order system' coefficient array that scipy filters generate when created

    Returns:
        A 2D array of the Time and GMR readings after empty plate calibration. Data will be rounded to integers if calibration results in slight decimal behavior.
    """
    time_readings = fully_calibrated_gmr[0, :]
    gmr_readings = fully_calibrated_gmr[1, :]

    float_array = signal.sosfiltfilt(scipy_filter_sos_coefficients, gmr_readings)
    int_array = np.rint(float_array).astype(np.int32)

    filtered_data: NDArray[(2, Any), int] = np.vstack((time_readings, int_array))
    return filtered_data


def calculate_voltage_from_gmr(
    gmr_data: NDArray[(2, Any), int],
    reference_voltage: Union[float, int] = 3.3,
    adc_gain: int = 32,
) -> NDArray[(2, Any), np.float32]:
    """Convert 'signed' 24-bit values from an ADC to measured voltage.

    Args:
        gmr_data: time and GMR numpy array. Typically coming from filtered_gmr_data
        reference_voltage: Almost always leave as default of 3.3V
        adc_gain: Current implementation of Mantarray is constant value of 32, but may change in the future. Value can be obtained from HDF5 metadata

    Returns:
        A 2D array of time vs Voltage
    """
    voltage = (
        gmr_data[1, :].astype(np.float32)
        * (1 / RAW_TO_SIGNED_CONVERSION_VALUE)
        * (reference_voltage / adc_gain)
    )
    return np.vstack((gmr_data[0, :].astype(np.float32), voltage))


def calculate_displacement_from_voltage(
    voltage_data: NDArray[(2, Any), np.float32],
) -> NDArray[(2, Any), np.float32]:
    """Convert voltage to displacement.

    Args:
        voltage_data: time and Voltage numpy array. Typically coming from calculate_voltage_from_gmr

    Returns:
        A 2D array of time vs Displacement
    """
    voltage = voltage_data[1, :]
    time = voltage_data[0, :]
    displacement = voltage * -1 + 0.1  # simplistically just invert the data for now
    return np.vstack((time, displacement)).astype(np.float32)



# cython: language_level=3
# Tanner (9/1/20): Make sure to set `linetrace=False` except when profiling cython code or creating annotation file. All performance tests should be timed without line tracing enabled. Cython files in this package can easily be recompiled with `pip install -e .`
# cython: linetrace=False
# Eli (8/18/20) ... not sure why cython doesn't default to compiling with Python 3...but apparently this is explicitly needed https://github.com/cython/cython/issues/2299
"""Compressions arrays of Mantarray GMR data ."""
from typing import Any

from nptyping import NDArray
import numpy as np

cdef float R_SQUARE_CUTOFF = 0.94

cpdef float rsquared(int[:] x_values, int[:] y_values):
    """Return R^2 where x and y are array-like.
    Instead of doing a full linear regression, the compression process drops all points in between the first and the last, so R^2 residuals are calculated based off of the line between the first and last points.
    Typically the time and filtered GMR readings are supplied as the x and y axis values respectively.
    Args:
        x_values: 1D array of int values representing time
        y_values: 1D array of int values representing filtered GMR
    Returns:
        the R^2 value of the given dataset
    """
    cdef int x_0, x_1, y_0, y_1
    cdef float slope, intercept, ss_res, ss_tot, y_bar


    x_0 = x_values[0]
    x_1 = x_values[-1]
    y_0 = y_values[0]
    y_1 = y_values[-1]
    slope = (y_1 - y_0) / (x_1 - x_0)
    intercept = -slope * x_1 + y_1

    # based on https://stackoverflow.com/questions/893657/how-do-i-calculate-r-squared-using-python-and-numpy
    cdef int y_sum, num_values, i
    y_sum=0
    num_values=y_values.shape[0]
    ss_res=0
    for i in range(num_values):
        y_sum += y_values[i]
        ss_res += (x_values[i] * slope + intercept - y_values[i]) ** 2
    y_bar = y_sum / num_values

    ss_tot = 0
    for i in range(num_values):
        ss_tot += (y_values[i] - y_bar) ** 2

    if ss_tot == 0:  # Tanner (8/31/20): If a flat, horizontal line is passed to this function, ss_tot will equal 0, so we must handle this edge case to avoid Div By Zero Errors
        return 1.0
    return 1 - ss_res / ss_tot


def compress_filtered_gmr(int[:,:] data) -> NDArray[(2, Any), int]:
    """Compress the data to allow for better plotting in the desktop app.

    Args:
        data: a 2D array of GMR data after noise filtering

    Returns:
        a 2D array containing slightly less points than the original data
    """
    # split time and GMR readings into individual arrays
    cdef int[:] time = data[0, :]
    cdef int[:] filtered_gmr = data[1, :]
    cdef int time_len = len(time)

    # create a boolean array of indicies that will be kept
    what_to_keep = [True] * time_len

    # loop through values in time and filtered_gmr to determine what to compress
    cdef int left_idx, right_idx
    cdef int[:] subset_time
    cdef int[:] subset_filtered_gmr

    left_idx = 0
    while left_idx < time_len - 2:
        right_idx = left_idx + 3  # create an initial subset of length 3

        subset_time = time[left_idx:right_idx]
        subset_filtered_gmr = filtered_gmr[left_idx:right_idx]

        # calculate r_squared value of the initial length 3 subset
        r_squared = rsquared(subset_time, subset_filtered_gmr)

        if r_squared > R_SQUARE_CUTOFF:
            while r_squared > R_SQUARE_CUTOFF and right_idx < time_len:
                what_to_keep[right_idx - 2] = False
                # add another point into the subset
                right_idx += 1
                subset_time = time[left_idx:right_idx]
                subset_filtered_gmr = filtered_gmr[left_idx:right_idx]

                # re-calculate the new r_squared
                r_squared = rsquared(subset_time, subset_filtered_gmr)
            if r_squared > R_SQUARE_CUTOFF and right_idx == time_len:
                what_to_keep[right_idx - 2] = False

            left_idx = right_idx - 1

        else:
            left_idx += 1

    # compress the arrays
    compressed_time = np.compress(what_to_keep, time)
    compressed_filtered_gmr = np.compress(what_to_keep, filtered_gmr)

    # combine the arrays back into proper format
    compressed_data = np.array([compressed_time, compressed_filtered_gmr], dtype=np.int32)

    return compressed_data


# -*- coding: utf-8 -*-
"""Detecting peak and valleys of incoming Mantarray data."""

from functools import partial
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from uuid import UUID

from nptyping import NDArray
import numpy as np
from scipy import signal

from .constants import AMPLITUDE_UUID
from .constants import AUC_UUID
from .constants import CENTIMILLISECONDS_PER_SECOND
from .constants import MIN_NUMBER_PEAKS
from .constants import PRIOR_PEAK_INDEX_UUID
from .constants import PRIOR_VALLEY_INDEX_UUID
from .constants import SUBSEQUENT_PEAK_INDEX_UUID
from .constants import SUBSEQUENT_VALLEY_INDEX_UUID
from .constants import TWITCH_FREQUENCY_UUID
from .constants import TWITCH_PERIOD_UUID
from .constants import WIDTH_FALLING_COORDS_UUID
from .constants import WIDTH_RISING_COORDS_UUID
from .constants import WIDTH_UUID
from .constants import WIDTH_VALUE_UUID
from .exceptions import TooFewPeaksDetectedError
from .exceptions import TwoPeaksInARowError
from .exceptions import TwoValleysInARowError

TWITCH_WIDTH_PERCENTS = range(10, 95, 5)


def peak_detector(
    filtered_gmr: NDArray[(2, Any), int], twitches_point_up: bool = True,
) -> Tuple[List[int], List[int]]:
    """Locates peaks and valleys and returns the indices.

    Args:
        noise_free_data: a 2D array of the time and voltage data after it has gone through noise cancellation
        sampling_rate: an integer value of the sampling rate of the data in Hz
        twitches_point_up: whether in the incoming data stream the biological twitches are pointing up (in the positive direction) or down
        data: a 2D array of the original time and voltage before noise cancellation

    Returns:
        A tuple of the indices of the peaks and valleys
    """
    gmr_signal: NDArray[int] = filtered_gmr[1, :]
    peak_invertor_factor = 1
    valley_invertor_factor = -1
    if not twitches_point_up:
        peak_invertor_factor *= -1
        valley_invertor_factor *= -1
    sampling_period_cms = filtered_gmr[0, 1] - filtered_gmr[0, 0]
    maximum_possible_twitch_frequency = 7  # pylint:disable=invalid-name # (Eli 9/1/20: I can't think of a shorter name to describe this concept fully) # Hz
    minimum_required_samples_between_twitches = int(  # pylint:disable=invalid-name # (Eli 9/1/20: I can't think of a shorter name to describe this concept fully)
        round(
            (1 / maximum_possible_twitch_frequency)
            * CENTIMILLISECONDS_PER_SECOND
            / sampling_period_cms,
            0,
        )
    )

    # find required height of peaks
    max_height = np.max(gmr_signal)
    min_height = np.min(gmr_signal)
    max_prominence = abs(max_height - min_height)
    # print(f"max prom {max_prominence}")
    # find peaks and valleys
    peak_indices, _ = signal.find_peaks(
        gmr_signal * peak_invertor_factor,
        width=minimum_required_samples_between_twitches / 2,
        distance=minimum_required_samples_between_twitches,
        prominence=max_prominence / 22,
    )
    valley_indices, _ = signal.find_peaks(
        gmr_signal * valley_invertor_factor,
        width=minimum_required_samples_between_twitches / 2,
        distance=minimum_required_samples_between_twitches,
        prominence=max_prominence / 22,
    )
    return peak_indices, valley_indices


def create_avg_dict(
    metric: NDArray[int], round_to_int: bool = True
) -> Dict[str, Union[float, int]]:
    """Calculate the average values of a specific metric.

    Args:
        metric: a 1D array of integer values of a specific metric results

    Returns:
        a dictionary of the average statistics of that metric
    """
    dictionary: Dict[str, Union[float, int]] = dict()

    dictionary["n"] = len(metric)
    dictionary["mean"] = np.mean(metric)
    dictionary["std"] = np.std(metric)
    dictionary["min"] = np.min(metric)
    dictionary["max"] = np.max(metric)
    if round_to_int:
        for iter_key in ("mean", "std", "min", "max"):
            dictionary[iter_key] = int(round(dictionary[iter_key]))

    return dictionary


def data_metrics(
    peak_and_valley_indices: Tuple[NDArray[int], NDArray[int]],
    filtered_data: NDArray[(2, Any), int],
) -> Tuple[
    Dict[int, Dict[UUID, Union[float, int]]],
    Dict[
        UUID,
        Union[Dict[str, Union[float, int]], Dict[int, Dict[str, Union[float, int]]]],
    ],
]:  # pylint:disable=too-many-locals # Eli (9/8/20): there are a lot of metrics to calculate that need local variables
    """Find all data metrics for individual twitches and averages.

    Args:
        peakind: a tuple of integer values representing the time indices of peaks and valleys within the data
        filtered_data: a 2D array of the time and voltage data after it has gone through noise cancellation

    Returns:
        per_twitch_dict: a dictionary of individual peak metrics
        aggregate_dict: a dictionary of entire metric statistics. Most metrics have the stats underneath the UUID, but for twitch widths, there is an additional dictionary where the percent of repolarization is the key
    """
    # create main dictionaries
    main_twitch_dict: Dict[int, Dict[UUID, Union[float, int]]] = dict()
    aggregate_dict: Dict[
        UUID,
        Union[Dict[str, Union[float, int]], Dict[int, Dict[str, Union[float, int]]]],
    ] = dict()

    # create dependent dicitonaries
    period_averages_dict: Dict[str, Union[float, int]] = {}
    amplitude_averages_dict: Dict[str, Union[float, int]] = {}
    auc_averages_dict: Dict[str, Union[float, int]] = {}

    peak_indices, _ = peak_and_valley_indices

    # find twitch time points

    twitch_indices = find_twitch_indices(peak_and_valley_indices, filtered_data)
    num_twitches = len(twitch_indices)
    time_series = filtered_data[0, :]

    # find twitch periods
    combined_twitch_periods = calculate_twitch_period(
        twitch_indices, peak_indices, filtered_data
    )

    twitch_frequencies = 1 / (
        combined_twitch_periods.astype(np.float) / CENTIMILLISECONDS_PER_SECOND
    )
    frequency_averages_dict = create_avg_dict(twitch_frequencies, round_to_int=False)
    # find aggregate values of period data
    period_averages_dict = create_avg_dict(combined_twitch_periods)

    aggregate_dict[TWITCH_PERIOD_UUID] = period_averages_dict
    aggregate_dict[TWITCH_FREQUENCY_UUID] = frequency_averages_dict

    # find twitch amplitudes
    amplitudes: NDArray[int] = calculate_amplitudes(twitch_indices, filtered_data)

    # find aggregate values of amplitude data
    amplitude_averages_dict = create_avg_dict(amplitudes)

    aggregate_dict[AMPLITUDE_UUID] = amplitude_averages_dict

    # find twitch widths
    widths = calculate_twitch_widths(twitch_indices, filtered_data)
    width_stats_dict: Dict[int, Dict[str, Union[float, int]]] = dict()

    for iter_percent in TWITCH_WIDTH_PERCENTS:
        iter_list_of_width_values: List[Union[float, int]] = []
        for iter_twitch in widths:
            iter_width_value = iter_twitch[iter_percent][WIDTH_VALUE_UUID]
            if not isinstance(iter_width_value, (float, int)):  # making mypy happy
                raise NotImplementedError(
                    f"The width value under key {WIDTH_VALUE_UUID} must be a float or an int. It was: {iter_width_value}"
                )
            iter_list_of_width_values.append(iter_width_value)
        iter_stats_dict = create_avg_dict(iter_list_of_width_values)
        width_stats_dict[iter_percent] = iter_stats_dict

    aggregate_dict[WIDTH_UUID] = width_stats_dict
    # calculate auc
    auc_per_twitch: NDArray[int] = calculate_area_under_curve(
        twitch_indices, filtered_data, widths
    )
    # find aggregate values of area under curve data
    auc_averages_dict = create_avg_dict(auc_per_twitch)

    aggregate_dict[AUC_UUID] = auc_averages_dict

    # add metrics to per peak dictionary
    twitch_peak_indices = tuple(twitch_indices.keys())
    for i in range(num_twitches):
        main_twitch_dict.update(
            {
                time_series[twitch_peak_indices[i]]: {
                    TWITCH_PERIOD_UUID: combined_twitch_periods[i],
                    AMPLITUDE_UUID: amplitudes[i],
                    WIDTH_UUID: widths[i],
                    AUC_UUID: auc_per_twitch[i],
                    TWITCH_FREQUENCY_UUID: twitch_frequencies[i],
                }
            }
        )

    return main_twitch_dict, aggregate_dict


def calculate_twitch_period(
    twitch_indices: NDArray[int],
    all_peak_indices: NDArray[int],
    filtered_data: NDArray[(2, Any), int],
) -> NDArray[int]:
    """Find the distance between each twitch at its peak.

    Args:
        twitch_indices: a 1D array of the indices in the data array that the twitch peaks are at
        all_peak_indices: a 1D array of the indices in teh data array that all peaks are at
        filtered_data: a 2D array (time vs value) of the data

    Returns:
        an array of integers that are the peeriod of each twitch
    """
    list_of_twitch_indices = list(twitch_indices.keys())
    idx_of_first_twitch = np.where(all_peak_indices == list_of_twitch_indices[0])[0][0]
    period: List[int] = []
    time_series = filtered_data[0, :]
    for iter_twitch_idx in range(len(list_of_twitch_indices)):

        period.append(
            time_series[all_peak_indices[iter_twitch_idx + idx_of_first_twitch + 1]]
            - time_series[all_peak_indices[iter_twitch_idx + idx_of_first_twitch]]
        )

    return np.asarray(period, dtype=np.int32)


def find_twitch_indices(
    peak_and_valley_indices: Tuple[NDArray[int], NDArray[int]],
    filtered_data: NDArray[(2, Any), int],
) -> Dict[int, Dict[UUID, Optional[int]]]:
    """Find twitches that can be analyzed.

    Sometimes the first and last peak in a trace can't be analyzed as a full twitch because not enough information is present.
    In order to be analyzable, a twitch needs to have a valley prior to it and another peak after it.

    Args:
        peak_and_valley_indices: a Tuple of 1D array of integers representing the indices of the peaks and valleys
        filtered_data: a 2D array of the data after being noise filtered

    Returns:
        a 1D array of integers representing the time points of all the twitches
    """
    peak_indices, valley_indices = peak_and_valley_indices
    if len(peak_indices) < MIN_NUMBER_PEAKS:
        raise TooFewPeaksDetectedError(
            f"A minimum of {MIN_NUMBER_PEAKS} peaks is required to extract twitch metrics, however only {len(peak_indices)} peak(s) were detected"
        )
    twitches: Dict[int, Dict[UUID, Optional[int]]] = {}
    starts_with_peak = peak_indices[0] < valley_indices[0]
    # ends_with_peak = peak_indices[-1] > valley_indices[-1]
    for itr_idx, itr_peak_index in enumerate(peak_indices):
        if itr_idx == peak_indices.shape[0] - 1:  # last peak
            if itr_idx == valley_indices.shape[0] - 2:
                raise TwoValleysInARowError(
                    peak_and_valley_indices,
                    filtered_data,
                    (valley_indices[itr_idx], valley_indices[itr_idx + 1]),
                )
            if itr_idx == valley_indices.shape[0] - 3:
                raise TwoValleysInARowError(
                    peak_and_valley_indices,
                    filtered_data,
                    (valley_indices[itr_idx + 1], valley_indices[itr_idx + 2]),
                )
        else:
            if starts_with_peak:
                if (
                    itr_idx == valley_indices.shape[0]
                    and itr_idx == peak_indices.shape[0] - 2
                    or valley_indices[itr_idx] > peak_indices[itr_idx + 1]
                ):
                    raise TwoPeaksInARowError(
                        peak_and_valley_indices,
                        filtered_data,
                        (peak_indices[itr_idx], peak_indices[itr_idx + 1]),
                    )
                if itr_peak_index > valley_indices[itr_idx]:
                    raise TwoValleysInARowError(
                        peak_and_valley_indices,
                        filtered_data,
                        (valley_indices[itr_idx - 1], valley_indices[itr_idx]),
                    )
            else:
                if (
                    itr_idx == valley_indices.shape[0] - 1
                    and itr_idx == peak_indices.shape[0] - 2
                ):
                    raise TwoPeaksInARowError(
                        peak_and_valley_indices,
                        filtered_data,
                        (peak_indices[itr_idx], peak_indices[itr_idx + 1]),
                    )
                if valley_indices[itr_idx] > peak_indices[itr_idx]:
                    raise TwoPeaksInARowError(
                        peak_and_valley_indices,
                        filtered_data,
                        (peak_indices[itr_idx - 1], peak_indices[itr_idx]),
                    )
                if (
                    itr_idx < len(valley_indices) - 1
                    and itr_peak_index > valley_indices[itr_idx + 1]
                ):
                    raise TwoValleysInARowError(
                        peak_and_valley_indices,
                        filtered_data,
                        (valley_indices[itr_idx], valley_indices[itr_idx + 1]),
                    )

            if itr_idx == 0 and starts_with_peak:
                continue

            twitches[itr_peak_index] = {
                PRIOR_PEAK_INDEX_UUID: None
                if itr_idx == 0
                else peak_indices[itr_idx - 1],
                PRIOR_VALLEY_INDEX_UUID: valley_indices[
                    itr_idx - 1 if starts_with_peak else itr_idx
                ],
                SUBSEQUENT_PEAK_INDEX_UUID: peak_indices[itr_idx + 1],
                SUBSEQUENT_VALLEY_INDEX_UUID: valley_indices[
                    itr_idx if starts_with_peak else itr_idx + 1
                ],
            }

    # print(list(twitches.keys())[0])
    # print(twitches[list(twitches.keys())[0]])
    return twitches


def calculate_amplitudes(
    twitch_indices: Dict[int, Dict[UUID, Optional[int]]],
    filtered_data: NDArray[(2, Any), int],
) -> NDArray[int]:
    """Get the amplitudes for all twitches.

    Args:
        twitch_indices: a 1D array of all the time values of the peaks of interest
        filtered_data: a 2D array of the time and value (magnetic, voltage, displacement, force...) data after it has gone through noise filtering

    Returns:
        a 1D array of integers representing the amplitude of each twitch
    """
    amplitudes: List[int] = list()
    amplitude_series = filtered_data[1, :]
    for iter_twitch_peak_idx, iter_twitch_indices_info in twitch_indices.items():
        peak_amplitude = amplitude_series[iter_twitch_peak_idx]
        prior_amplitude = amplitude_series[
            iter_twitch_indices_info[PRIOR_VALLEY_INDEX_UUID]
        ]
        subsequent_amplitude = amplitude_series[
            iter_twitch_indices_info[SUBSEQUENT_VALLEY_INDEX_UUID]
        ]
        amplitudes.append(
            abs(
                int(
                    round(
                        (
                            (peak_amplitude - prior_amplitude)
                            + (peak_amplitude - subsequent_amplitude)
                        )
                        / 2,
                        0,
                    )
                )
            )
        )

    return np.asarray(amplitudes, dtype=np.int32)


def interpolate_x_for_y_between_two_points(  # pylint:disable=invalid-name # (Eli 9/1/20: I can't think of a shorter name to describe this concept fully)
    desired_y: Union[int, float],
    x_1: Union[int, float],
    y_1: Union[int, float],
    x_2: Union[int, float],
    y_2: Union[int, float],
) -> Union[int, float]:
    """Find a value of x between two points that matches the desired y value.

    Uses linear interpolation, based on point-slope formula.
    """
    slope = (y_2 - y_1) / (x_2 - x_1)
    return (desired_y - y_1) / slope + x_1


def interpolate_y_for_x_between_two_points(  # pylint:disable=invalid-name # (Eli 9/1/20: I can't think of a shorter name to describe this concept fully)
    desired_x: Union[int, float],
    x_1: Union[int, float],
    y_1: Union[int, float],
    x_2: Union[int, float],
    y_2: Union[int, float],
) -> Union[int, float]:
    """Find a value of x between two points that matches the desired y value.

    Uses linear interpolation, based on point-slope formula.
    """
    slope = (y_2 - y_1) / (x_2 - x_1)
    return slope * (desired_x - x_1) + y_1


def calculate_twitch_widths(
    twitch_indices: Dict[int, Dict[UUID, Optional[int]]],
    filtered_data: NDArray[(2, Any), int],
) -> List[Dict[int, Dict[UUID, Union[Tuple[int, int], int]]]]:
    """Determine twitch width between 10-90% down to the nearby valleys.

    Args:
        twitch_indices: a 1D array of all the time values of the peaks of interest
        filtered_data: a 2D array of the time and value (magnetic, voltage, displacement, force...) data after it has gone through noise filtering

    Returns:
        a list of dictionaries where the first key is the percentage of the way down to the nearby valleys, the second key is a UUID representing either the value of the width, or the rising or falling coordinates. The final value is either an int (for value) or a tuple of ints for the x/y coordinates
    """
    widths: List[Dict[int, Dict[UUID, Union[Tuple[int, int], int]]]] = list()
    value_series = filtered_data[1, :]
    time_series = filtered_data[0, :]
    for iter_twitch_peak_idx, iter_twitch_indices_info in twitch_indices.items():
        iter_width_dict: Dict[int, Dict[UUID, Union[Tuple[int, int], int]]] = dict()
        peak_value = value_series[iter_twitch_peak_idx]
        prior_valley_value = value_series[
            iter_twitch_indices_info[PRIOR_VALLEY_INDEX_UUID]
        ]
        subsequent_valley_value = value_series[
            iter_twitch_indices_info[SUBSEQUENT_VALLEY_INDEX_UUID]
        ]

        rising_amplitude = peak_value - prior_valley_value
        falling_amplitude = peak_value - subsequent_valley_value

        rising_idx = iter_twitch_peak_idx - 1
        falling_idx = iter_twitch_peak_idx + 1
        for iter_percent in TWITCH_WIDTH_PERCENTS:
            iter_percent_dict: Dict[UUID, Union[Tuple[int, int], int]] = dict()
            rising_threshold = peak_value - iter_percent / 100 * rising_amplitude
            falling_threshold = peak_value - iter_percent / 100 * falling_amplitude
            # move to the left from the twitch peak until the threshold is reached
            while abs(value_series[rising_idx] - prior_valley_value) > abs(
                rising_threshold - prior_valley_value
            ):
                rising_idx -= 1
            # move to the right from the twitch peak until the falling threshold is reached
            while abs(value_series[falling_idx] - subsequent_valley_value) > abs(
                falling_threshold - subsequent_valley_value
            ):
                falling_idx += 1
            interpolated_rising_timepoint = interpolate_x_for_y_between_two_points(
                rising_threshold,
                time_series[rising_idx],
                value_series[rising_idx],
                time_series[rising_idx + 1],
                value_series[rising_idx + 1],
            )
            interpolated_falling_timepoint = interpolate_x_for_y_between_two_points(
                falling_threshold,
                time_series[falling_idx],
                value_series[falling_idx],
                time_series[falling_idx - 1],
                value_series[falling_idx - 1],
            )
            iter_percent_dict[WIDTH_VALUE_UUID] = int(
                round(interpolated_falling_timepoint - interpolated_rising_timepoint, 0)
            )
            iter_percent_dict[WIDTH_RISING_COORDS_UUID] = (
                int(round(interpolated_rising_timepoint, 0)),
                int(round(rising_threshold, 0)),
            )
            iter_percent_dict[WIDTH_FALLING_COORDS_UUID] = (
                int(round(interpolated_falling_timepoint, 0)),
                int(round(falling_threshold, 0)),
            )
            iter_width_dict[iter_percent] = iter_percent_dict
        widths.append(iter_width_dict)
    return widths


def calculate_area_under_curve(  # pylint:disable=too-many-locals # Eli (9/1/20): may be able to refactor before pull request
    twitch_indices: Dict[int, Dict[UUID, Optional[int]]],
    filtered_data: NDArray[(2, Any), int],
    per_twitch_widths: List[Dict[int, Dict[UUID, Union[Tuple[int, int], int]]]],
) -> NDArray[int]:
    """Calculate the area under the curve (AUC) for twitches."""
    width_percent = 90  # what percent of repolarization to use as the bottom limit for calculating AUC
    auc_per_twitch: List[int] = list()
    value_series = filtered_data[1, :]
    time_series = filtered_data[0, :]

    for iter_twitch_idx, (iter_twitch_peak_idx, iter_twitch_indices_info) in enumerate(
        twitch_indices.items()
    ):
        # iter_twitch_peak_timepoint = time_series[iter_twitch_peak_idx]
        width_info = per_twitch_widths[iter_twitch_idx]
        prior_valley_value = value_series[
            iter_twitch_indices_info[PRIOR_VALLEY_INDEX_UUID]
        ]
        subsequent_valley_value = value_series[
            iter_twitch_indices_info[SUBSEQUENT_VALLEY_INDEX_UUID]
        ]
        rising_coords = width_info[width_percent][WIDTH_RISING_COORDS_UUID]
        if not isinstance(
            rising_coords, tuple
        ):  # Eli (9/1/20): this appears needed to make mypy happy
            raise NotImplementedError(
                f"Rising coordinates under the key {WIDTH_RISING_COORDS_UUID} must be a tuple."
            )
        rising_x, rising_y = rising_coords
        falling_coords = width_info[width_percent][WIDTH_FALLING_COORDS_UUID]
        if not isinstance(
            falling_coords, tuple
        ):  # Eli (9/1/20): this appears needed to make mypy happy
            raise NotImplementedError(
                f"Falling coordinates under the key {WIDTH_FALLING_COORDS_UUID} must be a tuple."
            )
        falling_x, falling_y = falling_coords
        interp_y_for_lower_bound = partial(
            interpolate_y_for_x_between_two_points,
            x_1=rising_x,
            y_1=rising_y,
            x_2=falling_x,
            y_2=falling_y,
        )
        auc_total: Union[float, int] = 0

        # calculate area of rising side
        rising_idx = iter_twitch_peak_idx
        # move to the left from the twitch peak until the threshold is reached
        while abs(value_series[rising_idx - 1] - prior_valley_value) > abs(
            rising_y - prior_valley_value
        ):
            left_x = time_series[rising_idx - 1]
            right_x = time_series[rising_idx]
            left_y = value_series[rising_idx - 1]
            right_y = value_series[rising_idx]
            trapezoid_h = right_x - left_x
            trapezoid_left_side = abs(left_y - interp_y_for_lower_bound(left_x))
            trapezoid_right_side = abs(right_y - interp_y_for_lower_bound(right_x))

            auc_total += (trapezoid_left_side + trapezoid_right_side) / 2 * trapezoid_h
            rising_idx -= 1
        # final trapezoid at the boundary of the interpolated twitch width point
        left_x = rising_x
        right_x = time_series[rising_idx]
        left_y = rising_y
        right_y = value_series[rising_idx]
        trapezoid_h = right_x - left_x
        trapezoid_left_side = abs(left_y - interp_y_for_lower_bound(left_x))
        trapezoid_right_side = abs(right_y - interp_y_for_lower_bound(right_x))
        auc_total += (trapezoid_left_side + trapezoid_right_side) / 2 * trapezoid_h

        # calculate area of falling side
        falling_idx = iter_twitch_peak_idx
        # move to the left from the twitch peak until the threshold is reached
        while abs(value_series[falling_idx + 1] - subsequent_valley_value) > abs(
            falling_y - subsequent_valley_value
        ):
            left_x = time_series[falling_idx]
            right_x = time_series[falling_idx + 1]
            left_y = value_series[falling_idx]
            right_y = value_series[falling_idx + 1]
            trapezoid_h = right_x - left_x
            trapezoid_left_side = abs(left_y - interp_y_for_lower_bound(left_x))
            trapezoid_right_side = abs(right_y - interp_y_for_lower_bound(right_x))
            auc_total += (trapezoid_left_side + trapezoid_right_side) / 2 * trapezoid_h
            falling_idx += 1

        # final trapezoid at the boundary of the interpolated twitch width point
        left_x = time_series[falling_idx]
        right_x = falling_x
        left_y = value_series[rising_idx]
        right_y = falling_y
        trapezoid_h = right_x - left_x
        trapezoid_left_side = abs(left_y - interp_y_for_lower_bound(left_x))
        trapezoid_right_side = abs(right_y - interp_y_for_lower_bound(right_x))
        auc_total += (trapezoid_left_side + trapezoid_right_side) / 2 * trapezoid_h

        auc_per_twitch.append(int(round(auc_total, 0)))

    return np.asarray(auc_per_twitch, dtype=np.int64)



# -*- coding: utf-8 -*-
"""Docstring."""
from . import pipelines
from . import transforms
from .constants import AMPLITUDE_UUID
from .constants import AUC_UUID
from .constants import BESSEL_BANDPASS_UUID
from .constants import BESSEL_LOWPASS_10_UUID
from .constants import BESSEL_LOWPASS_30_UUID
from .constants import BUTTERWORTH_LOWPASS_30_UUID
from .constants import CENTIMILLISECONDS_PER_SECOND
from .constants import MIDSCALE_CODE
from .constants import MIN_NUMBER_PEAKS
from .constants import PRIOR_PEAK_INDEX_UUID
from .constants import PRIOR_VALLEY_INDEX_UUID
from .constants import RAW_TO_SIGNED_CONVERSION_VALUE
from .constants import SUBSEQUENT_PEAK_INDEX_UUID
from .constants import SUBSEQUENT_VALLEY_INDEX_UUID
from .constants import TWITCH_FREQUENCY_UUID
from .constants import TWITCH_PERIOD_UUID
from .constants import WIDTH_FALLING_COORDS_UUID
from .constants import WIDTH_RISING_COORDS_UUID
from .constants import WIDTH_UUID
from .constants import WIDTH_VALUE_UUID
from .exceptions import DataAlreadyLoadedInPipelineError
from .exceptions import FilterCreationNotImplementedError
from .exceptions import TooFewPeaksDetectedError
from .exceptions import TwoPeaksInARowError
from .exceptions import TwoValleysInARowError
from .exceptions import UnrecognizedFilterUuidError
from .peak_detection import find_twitch_indices
from .peak_detection import peak_detector
from .pipelines import Pipeline
from .pipelines import PipelineTemplate
from .transforms import apply_empty_plate_calibration
from .transforms import apply_noise_filtering
from .transforms import apply_sensitivity_calibration
from .transforms import calculate_displacement_from_voltage
from .transforms import calculate_voltage_from_gmr
from .transforms import create_filter
from .transforms import FILTER_CHARACTERISTICS
from .transforms import noise_cancellation

if (
    5 < 10
):  # pragma: no cover # protect this from zimports deleting the pylint disable statement
    from .compression_cy import (  # pylint: disable=import-error # Eli (8/18/20) unsure why pylint is unable to recognize cython import... # Tanner (8/31/20) Pylint also flags this as duplicate code due to a similar import in pipelines.py, which may be related to pylint's import issues
        compress_filtered_gmr,
    )

__all__ = [
    "transforms",
    "pipelines",
    "TWITCH_PERIOD_UUID",
    "TWITCH_FREQUENCY_UUID",
    "PRIOR_PEAK_INDEX_UUID",
    "PRIOR_VALLEY_INDEX_UUID",
    "SUBSEQUENT_PEAK_INDEX_UUID",
    "SUBSEQUENT_VALLEY_INDEX_UUID",
    "AMPLITUDE_UUID",
    "AUC_UUID",
    "WIDTH_UUID",
    "WIDTH_VALUE_UUID",
    "WIDTH_RISING_COORDS_UUID",
    "WIDTH_FALLING_COORDS_UUID",
    "CENTIMILLISECONDS_PER_SECOND",
    "MIDSCALE_CODE",
    "RAW_TO_SIGNED_CONVERSION_VALUE",
    "apply_sensitivity_calibration",
    "noise_cancellation",
    "apply_empty_plate_calibration",
    "apply_noise_filtering",
    "create_filter",
    "UnrecognizedFilterUuidError",
    "FilterCreationNotImplementedError",
    "DataAlreadyLoadedInPipelineError",
    "TwoPeaksInARowError",
    "TwoValleysInARowError",
    "BESSEL_BANDPASS_UUID",
    "BESSEL_LOWPASS_10_UUID",
    "FILTER_CHARACTERISTICS",
    "compress_filtered_gmr",
    "calculate_voltage_from_gmr",
    "calculate_displacement_from_voltage",
    "PipelineTemplate",
    "Pipeline",
    "peak_detector",
    "find_twitch_indices",
    "TooFewPeaksDetectedError",
    "MIN_NUMBER_PEAKS",
    "BESSEL_LOWPASS_30_UUID",
    "BUTTERWORTH_LOWPASS_30_UUID",
]



# -*- coding: utf-8 -*-
"""Setup configuration."""
import os

from setuptools import Extension
from setuptools import find_packages
from setuptools import setup

# In order to locally Cythonize files, make sure to have installed the Python dev module: sudo apt-get install python3.7-dev

try:
    from Cython.Build import cythonize
except ImportError:
    USE_CYTHON = False
else:
    USE_CYTHON = True

ext = ".pyx" if USE_CYTHON else ".c"
extensions = [
    Extension(
        "mantarray_waveform_analysis.compression_cy",
        [os.path.join("src", "mantarray_waveform_analysis", "compression_cy") + ext],
    )
]

if USE_CYTHON:
    # cythonizing compression_cy.pyx with kwarg annotate=True will help when optimizing the code by enabling generation of the html annotation file
    extensions = cythonize(extensions, annotate=False)


setup(
    name="mantarray_waveform_analysis",
    version="0.5.3",
    description="Tools for analyzing waveforms produced by a Mantarray Instrument",
    url="https://github.com/CuriBio/mantarray-waveform-analysis",
    author="Curi Bio",
    author_email="contact@curibio.com",
    license="MIT",
    packages=find_packages("src"),
    package_dir={"": "src"},
    install_requires=[
        "numpy>=1.19.1",
        "scipy>=1.5.2",
        "nptyping>=1.3.0",
        "attrs>=19.3.0",
    ],
    zip_safe=False,
    include_package_data=True,
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
    ],
    ext_modules=extensions,
)



# -*- coding: utf-8 -*-
"""Transforming arrays of Mantarray data throughout the analysis pipeline."""
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from uuid import UUID

import attr
from nptyping import NDArray
import numpy as np

from .exceptions import DataAlreadyLoadedInPipelineError
from .peak_detection import data_metrics
from .peak_detection import peak_detector
from .transforms import apply_empty_plate_calibration
from .transforms import apply_noise_filtering
from .transforms import apply_sensitivity_calibration
from .transforms import calculate_displacement_from_voltage
from .transforms import calculate_voltage_from_gmr
from .transforms import create_filter
from .transforms import noise_cancellation

if (
    6 < 9
):  # pragma: no cover # protect this from zimports deleting the pylint disable statement
    from .compression_cy import (  # pylint: disable=import-error # Eli (8/18/20) unsure why pylint is unable to recognize cython import...
        compress_filtered_gmr,
    )


class Pipeline:
    """A pipeline used for analysis of a set of data.

    Args:
        pipeline_template: The PipelineTemplate instance that should be used to control the settings and operation of this analysis pipeline.
    """

    # pylint:disable=too-many-instance-attributes # Eli (9/8/20): it's a pipeline with lots of stages
    def __init__(self, pipeline_template: "PipelineTemplate") -> None:
        self._pipeline_template = pipeline_template
        self._raw_tissue_magnetic_data: NDArray[(2, Any), int]
        self._raw_reference_magnetic_data: NDArray[(2, Any), int]
        self._sensitivity_calibrated_tissue_gmr: NDArray[(2, Any), int]
        self._sensitivity_calibrated_reference_gmr: NDArray[(2, Any), int]
        self._noise_cancelled_magnetic_data: NDArray[(2, Any), int]
        self._fully_calibrated_magnetic_data: NDArray[(2, Any), int]
        self._noise_filtered_magnetic_data: NDArray[(2, Any), int]
        self._compressed_magnetic_data: NDArray[(2, Any), int]
        self._compressed_voltage: NDArray[(2, Any), np.float32]
        self._compressed_displacement: NDArray[(2, Any), np.float32]
        self._peak_detection_results: Tuple[List[int], List[int]]
        self._magnetic_data_metrics: Tuple[
            Dict[int, Dict[UUID, Union[float, int]]],
            Dict[
                UUID,
                Union[
                    Dict[str, Union[float, int]],
                    Dict[int, Dict[str, Union[float, int]]],
                ],
            ],
        ]

    def get_template(self) -> "PipelineTemplate":
        return self._pipeline_template

    def load_raw_magnetic_data(
        self,
        raw_tissue_magnetic_data: NDArray[(2, Any), int],
        raw_reference_magnetic_data: NDArray[(2, Any), int],
    ) -> None:
        """Load the raw magnetic sensor numpy arrays into the pipeline.

        Args:
            raw_tissue_magnetic_data: 2D array of time (in centimilliseconds) and magnetic reading for the sensor under the tissue construct
            raw_reference_magnetic_data: 2D array of time (in centimilliseconds) and magnetic reading for the sensor nearby the tissue construct
        """
        try:
            self._raw_tissue_magnetic_data
        except AttributeError:
            self._raw_tissue_magnetic_data = raw_tissue_magnetic_data
            self._raw_reference_magnetic_data = raw_reference_magnetic_data
            return
        raise DataAlreadyLoadedInPipelineError()

    def load_raw_gmr_data(
        self,
        raw_tissue_gmr: NDArray[(2, Any), int],
        raw_reference_gmr: NDArray[(2, Any), int],
    ) -> None:
        """Call magnetic data using this deprecated alias."""
        return self.load_raw_magnetic_data(raw_tissue_gmr, raw_reference_gmr)

    def get_raw_tissue_magnetic_data(self) -> NDArray[(2, Any), int]:
        return self._raw_tissue_magnetic_data

    def get_raw_reference_magnetic_data(self) -> NDArray[(2, Any), int]:
        return self._raw_reference_magnetic_data

    def get_sensitivity_calibrated_tissue_gmr(self) -> NDArray[(2, Any), int]:
        try:
            return self._sensitivity_calibrated_tissue_gmr
        except AttributeError:
            pass
        self._sensitivity_calibrated_tissue_gmr = apply_sensitivity_calibration(
            self._raw_tissue_magnetic_data
        )
        return self._sensitivity_calibrated_tissue_gmr

    def get_sensitivity_calibrated_reference_gmr(self) -> NDArray[(2, Any), int]:
        try:
            return self._sensitivity_calibrated_reference_gmr
        except AttributeError:
            pass
        self._sensitivity_calibrated_reference_gmr = apply_sensitivity_calibration(
            self._raw_reference_magnetic_data
        )
        return self._sensitivity_calibrated_reference_gmr

    def get_noise_cancelled_gmr(self) -> NDArray[(2, Any), int]:
        try:
            return self._noise_cancelled_magnetic_data
        except AttributeError:
            pass
        self._noise_cancelled_magnetic_data = noise_cancellation(
            self.get_sensitivity_calibrated_tissue_gmr(),
            self.get_sensitivity_calibrated_reference_gmr(),
        )
        return self._noise_cancelled_magnetic_data

    def get_fully_calibrated_gmr(self) -> NDArray[(2, Any), int]:
        try:
            return self._fully_calibrated_magnetic_data
        except AttributeError:
            pass
        self._fully_calibrated_magnetic_data = apply_empty_plate_calibration(
            self.get_noise_cancelled_gmr()
        )
        return self._fully_calibrated_magnetic_data

    def get_noise_filtered_gmr(self) -> NDArray[(2, Any), int]:
        return self.get_noise_filtered_magnetic_data()

    def get_noise_filtered_magnetic_data(self) -> NDArray[(2, Any), int]:
        """Return data after user-acceptable noise filtering."""
        try:
            return self._noise_filtered_magnetic_data
        except AttributeError:
            pass
        if self._pipeline_template.noise_filter_uuid is None:
            self._noise_filtered_magnetic_data = self.get_fully_calibrated_gmr()
        else:
            self._noise_filtered_magnetic_data = apply_noise_filtering(
                self.get_fully_calibrated_gmr(),
                self._pipeline_template.get_filter_coefficients(),
            )
        return self._noise_filtered_magnetic_data

    def get_peak_detection_results(self) -> Tuple[List[int], List[int]]:
        """Return peak detection results on noise filtered magnetic data."""
        try:
            return self._peak_detection_results
        except AttributeError:
            pass
        self._peak_detection_results = peak_detector(
            self.get_noise_filtered_magnetic_data(),
            twitches_point_up=self._pipeline_template.magnetic_twitches_point_up,
        )
        return self._peak_detection_results

    def get_magnetic_data_metrics(
        self,
    ) -> Tuple[
        Dict[int, Dict[UUID, Union[float, int]]],
        Dict[
            UUID,
            Union[
                Dict[str, Union[float, int]], Dict[int, Dict[str, Union[float, int]]]
            ],
        ],
    ]:
        """Calculate data metrics on noise filtered magnetic data."""
        try:
            return self._magnetic_data_metrics
        except AttributeError:
            pass
        self._magnetic_data_metrics = data_metrics(
            self.get_peak_detection_results(), self.get_noise_filtered_magnetic_data()
        )
        return self._magnetic_data_metrics

    def get_compressed_gmr(self) -> NDArray[(2, Any), int]:
        return self.get_compressed_magnetic_data()

    def get_compressed_magnetic_data(self) -> NDArray[(2, Any), int]:
        try:
            return self._compressed_magnetic_data
        except AttributeError:
            pass
        self._compressed_magnetic_data = compress_filtered_gmr(
            self.get_noise_filtered_magnetic_data()
        )
        return self._compressed_magnetic_data

    def get_compressed_voltage(self) -> NDArray[(2, Any), np.float32]:
        try:
            return self._compressed_voltage
        except AttributeError:
            pass
        self._compressed_voltage = calculate_voltage_from_gmr(self.get_compressed_gmr())
        return self._compressed_voltage

    def get_compressed_displacement(self) -> NDArray[(2, Any), np.float32]:
        try:
            return self._compressed_displacement
        except AttributeError:
            pass
        self._compressed_displacement = calculate_displacement_from_voltage(
            self.get_compressed_voltage()
        )
        return self._compressed_displacement


@attr.s
class PipelineTemplate:  # pylint: disable=too-few-public-methods # This is a simple dataclass/attrs module with just a few methods for validation and pipeline creation
    """A template the defines how an analysis pipeline should run.

    Args:
        noise_filter_uuid: The UUID corresponding to the filter settings that should be used when computationally filtering noise in the data.
        tissue_sampling_period: the sampling period for the tissues, in centimilliseconds
    """

    tissue_sampling_period: int = attr.ib()
    noise_filter_uuid: Optional[UUID] = attr.ib(default=None)
    magnetic_twitches_point_up: bool = attr.ib(default=False)
    _filter_coefficients: NDArray[(Any, Any), float]

    def create_pipeline(self) -> Pipeline:
        return Pipeline(self)

    def get_filter_coefficients(self) -> NDArray[(Any, Any), float]:
        """Get the coefficients for the signal filter.

        Creating a filter can take a non-trivial amount of time, so the
        coefficients are cached after the first time they are generated
        (since they are immutable within the template)
        """
        if self.noise_filter_uuid is None:
            raise NotImplementedError("Cannot create a filter when no UUID is set.")
        try:
            return self._filter_coefficients
        except AttributeError:
            pass
        self._filter_coefficients = create_filter(
            self.noise_filter_uuid, self.tissue_sampling_period
        )
        return self._filter_coefficients
